/*
   Intelligent Antenna Switch for Arduino

   This sketch will decode CAT frequency information
   from ICOM CI-V and switch antennas according
   to preset (memorized) values.

   Instead of antennas, bandfilters band will be selected

   Reads QRG from ICOM CI-V
   Determines band from readed QRG
   Stores prefered antenna for each of the
   changes the antenna when new freqency is in another band.

   This version can use any hardware serial port or software serial.
   Connect the RX pin to the CAT output through a 4k7 resistor.

   In this Version RX and TX Antenna are allways the same

   IMPORTANT : select on the rig in the menu
      CAT 'TRANSCEIVE' (ON)
   to implement broadcast of displayed frequency!

   to be done / backlog for improvements
   - improve button handling
   - (handle tx Antenna separately, yes/no)
   - only accept "from" device  / filter

   credits:
   This code is based on the idea of ON7EQ from November 2011
   https://www.qsl.net/on7eq/projects/arduino_ant_matrix.htm

   With thx to Jeff Smith VE1ZAC from which some code was inspired.
   Modified by Pete (El Supremo)

   project description:
   English: https://werner.rothschopf.net/microcontroller/202102_arduino_antennaswitch_en.htm
   German:  https://werner.rothschopf.net/microcontroller/202102_arduino_antennenschalter.htm

   by noiasca

   Version
   2023-08-14 new interface for YAESU RS232 and YAESU Linear 15pin socket, minor improvements
   2022-11-26 read buttons with digitalRead; example configuration for ESP32
   2022-07-03 new interface: read band information from ICOM ACC pin 5 (green)
   2022-06-04 optional filter from and filter to, watchdog
   2022-05-29 minor cleanup, fix offline indicator, improved package testing
   2022-03-27 display old style fixed warning, improved configuration
   2022-02-19 configuration in separate files
   2021-12-27 minor fix
   2021-11-05 process 100MHz - 9GHz
   2021-08-11 Changes for Noiasca Liquid Crystal 2.0.1
   2021-02-20 ok - compiles without errors/warnings on Arduino IDE 1.8.13
   2021-02-20 ok - first version for "Lois"
*/

#include <EEPROM.h>             // needed to store settings in EEPROM

#if defined(__AVR_ATmega328P__) || defined(__AVR_ATmega2560__)
#include <avr/wdt.h>            // Watchdog - only on Uno/Nano/Pro Mini/Mega
#endif

/* *************************************************************************
   structures 
   used in the configuration therefore needed before the #include of the configXXX.h
   ************************************************************************* */
struct Antenna {                                 // the structure for all antenna data
  const uint8_t pin;                             // GPIO of this antenna
  const char name[9];                            // name of antenna, you can use up to 8 characters (one less than the char size!)
};

struct Bandinformation {                         // this structure defines the frequence boundaries of the bands
  const uint32_t low;                            // lowest frequence in kilohertz (upgraded to 32bit to overcome 65.535 Limit)
  const uint32_t high;                           // highest frequence of this band
  const char name[5];                            // a clear text name for the band, remember you need one \0 as null terminator!
};

struct AdcBandinformation {                      // this structure defines the frequence boundaries of the bands
  const uint16_t adcLow;                         // lowest value on ADC
  const uint16_t adcHigh;                        // highest value on ADC
  const uint32_t qrg;                            // QRG within this band
};

struct LinearBandInformation {                   // this structure defines the frequence boundaries of the bands
  const uint8_t bitmap;                          // a 4 bit bitmap
  const uint32_t qrg;                            // QRG within this band
};

/* *************************************************************************
   program configuration is done with separate config files.
   copy/paste a configurartion tab, rename it and include ONE configuration file
   ************************************************************************* */
#include "configA.h"

/* *************************************************************************
   further settings,
   global variables and constants
   here you should only change things if you are very confident in Arduino programming
   ************************************************************************* */

int8_t antennaRX = 0;                            // current RX antenna (negative - no antenna connected)
// int8_t antennaTX = 0;                         // current TX antenna - not used in this version
const uint16_t readInterval = 500;               // some delay for the reading of button presses
uint32_t QRG = 0;                                // current frequency
int8_t band = -1;                                // the current band index we are on, must be signed for non valid/unknown bands
uint32_t previousQRGreceived = 0;                // timestamp of last received QRG

// a LCD with an PCF8574 expander
#include <Wire.h>                                // needed for I2C display
#include <NoiascaLiquidCrystal.h>                // Improved LCD I2C library: download from https://werner.rothschopf.net/202009_arduino_liquid_crystal_intro.htm 
#include <NoiascaHW/lcd_PCF8574.h>               // part of I2C lib: the I2C PCF8574 adapter
LiquidCrystal_PCF8574 lcd(lcdAddr, 16, 2);       // create LCD object

//as alternative, you can use a parallel 4bit LCD
//#include <NoiascaLiquidCrystal.h>              // Improved LCD I2C library: download from https://werner.rothschopf.net/202009_arduino_liquid_crystal_intro.htm
//#include <NoiascaHW/lcd_4bit.h>                // parallel interface, 4bit
//LiquidCrystal_4bit lcd(8, 9, 4,  5,  6,  7,  10, 16, 2);     // rs, en, d4, d5, d6, d7, bl, cols, rows - for keypad shield - double check against used pins!
//LiquidCrystal_4bit lcd(7, 8, 9, 10, 11, 12, 255, 16, 2);     // like in the ON7EQ sketch - double check against used pins!

// special characters for LCD in PROGMEM (not SRAM)
const uint8_t cat[8] PROGMEM =     {0b00000, 0b00000, 0b01110, 0b01000, 0b01110, 0b00000, 0b00000, 0b00000};    // QRG change
const uint8_t memory[8] PROGMEM =  {0b00000, 0b00000, 0b11011, 0b10101, 0b10001, 0b00000, 0b00000, 0b00000};    // memory activity
const uint8_t offline[8] PROGMEM = {0b00100, 0b00100, 0b11111, 0b00000, 0b00000, 0b11111, 0b00100, 0b00100};    // offline / no CAT for longer time
//const uint8_t error[8] PROGMEM =   {0b00000, 0b00000, 0b01010, 0b00100, 0b01010, 0b00000, 0b00000, 0b00000};    // error

#if USE_ADC
uint16_t previousAdc = 0;                        // the last read value from ACC port
#endif

#if USE_SERIAL == 0
HardwareSerial &mySerial = Serial;               // Reference to HW Serial (good for debugging of messages)
#elif USE_SERIAL == 1
HardwareSerial &mySerial = Serial1;              // Reference to HW Serial1 (for example if you are using an MEGA)
#elif USE_SERIAL == 2
HardwareSerial &mySerial = Serial2;
#elif USE_SERIAL == 3
HardwareSerial &mySerial = Serial3;
#else
#include <SoftwareSerial.h>
SoftwareSerial mySerial(catRxPin, catTxPin);     // RX, TX   // currently no CI-V messages are sent to mySerial, but for YAESU
#endif

#ifdef DEBUG_PRINT           // might be already definied in one of the libraries
#undef DEBUG_PRINT
#endif
#ifdef DEBUG_PRINTLN
#undef DEBUG_PRINTLN
#endif
#ifdef DEBUG
#define DEBUG_PRINT(...)    Serial.print (__VA_ARGS__)
#define DEBUG_PRINTLN(...)  Serial.println (__VA_ARGS__)
#else
#define DEBUG_PRINT(...)
#define DEBUG_PRINTLN(...)
#endif

// some functions and classes for the display/LCD
// if you change from a LCD to another type of display you have to modify this functions/class also

/*
   all three indicator react very similar - therefore a class
*/
class Indicator {
  private:
    uint32_t previousMillis;                     // last time shown
    bool active = false;                         // indicator is shown
    const uint8_t character;                     // the character to be displayed
    const uint8_t col;                           // on which column
    const uint8_t row;                           // on which row
    const uint16_t interval;                     // how long visible (after that time, the indicator will disapear)
  public:
    Indicator(uint8_t character, uint8_t col, uint8_t row, uint16_t interval = 512) :
      character(character), col(col), row(row), interval(interval) {}
    void on() {
      if (active == false) {
        lcd.setCursor(col, row);
        lcd.write(character);
        previousMillis = millis();
      }
      active = true;
    }
    void off() {
      if (active == true) {
        lcd.setCursor(col, row);
        lcd.write(' ');
      }
      active = false;
    }
    void update() {                              // to be called in loop()
      if (active == true && interval > 0) {
        if (millis() - previousMillis > interval) {
          off();
        }
      }
    }
};

Indicator indicatorActivity('-', 15, 0);
Indicator indicatorQRGchange(0, 15, 0);          // small c at the end of the first row
Indicator indicatorEEPROMchange(3, 15, 1);       // small m at the end of the second row
Indicator indicatorOffline(4, 15, 0, 0);         // no timeout --> 0 = permanent on

/*
   update antenna information on display
*/
void displayShowAntenna() {
  DEBUG_PRINT(F("D206 antennaRX=")); DEBUG_PRINTLN(antennaRX);
  lcd.setCursor(5, 1);
  lcd.print(antenna[antennaRX].name);
  lcd.setCursor(14, 1);
  lcd.print(antennaRX + 1);                      // +1 just because people don't understand why programmers start counting at 0...
}

/*
   update antenna information on display
   old style - only up to 5 Antennas  1 2(3)4 5
   use default brackets instead of custom characters
*/
void displayShowAntennaOld() {
  //uint8_t maxAntenna = min(5, noOfAntenna);       // we have only space for 5 antennas in this style
  uint8_t maxAntenna = noOfAntenna <= 5 ? noOfAntenna : 5;    // we have only space for 5 antennas in this style
  for (uint8_t i = 0; i < maxAntenna; i++) {
    lcd.write(' ');
    lcd.print(i + 1);                            // + 1 just because programmers start counting at ;-)
  }
  lcd.write(' ');
  if (antennaRX < maxAntenna) {
    lcd.setCursor((4 + antennaRX * 2), 1);
    lcd.write('(');
    lcd.setCursor((4 + antennaRX * 2 + 2), 1);
    lcd.write(')');
  }
}

/*
    show Band on display
    occupies 4 characters (and one trailing space)
    no parameter (or 0): show band according to global variable band
    paraemter != 0: show TX indicator
*/
void displayShowBand(uint8_t informationtype = 0) {
  lcd.setCursor(0, 1);                           // one common position for several information
  if (informationtype == 0) {
    if (band >= 0)
      lcd.print(tblBandinformation[band].name);
    else
      lcd.print(F("---"));
  }
  else {
    lcd.print(F("TX "));
  }
}

/*
  write frequence information to LCD
  Frequenz und Band information an LCD senden
*/
void displayShowQRG() {
#if USE_ADC
  // this prints the previous readed ADC raw value on the display
  // it makes debugging a little easier
  // you can comment the four lines if you prefere the fix text "QRG"
  lcd.setCursor(0, 0);
  lcd.print(F("    "));
  lcd.setCursor(0, 0);
  lcd.print(previousAdc);
#endif
  lcd.setCursor(4, 0);
  float output = QRG / 10000.0;
  if (output < 1) {
    lcd.print(output * 1000, 1);  // QRG is QRG*100, therefore one decimal, two blanks
    lcd.print(F("   KHz"));
  }
  else if (output < 1000) {
    if (output < 10) lcd.print(' ');
    if (output < 100)
      lcd.print ( output, 4);
    else
      lcd.print ( output, 3);
    lcd.print(F(" MHz"));
  }
  else {
    output = output / 1000;
    lcd.print (output, 5);
    lcd.print(F(" GHz"));
  }
  // previousDisplay = millis();
  displayShowBand();
}

/*
   on startup show configuration on LCD
*/
void displayShowStart() {
  lcd.setCursor(0, 0);
#if USE_CIV
  //           0123456789ABCDEF
  lcd.print(F("   Icom CI-V"));
#endif
#if USE_ADC
  lcd.print(F("   Icom ACC"));
#endif
#if USE_YAESU_RS232
  lcd.print(F("YAESU CAT RS232"));
#endif
#if USE_YAESU_LINEAR
  lcd.print(F("YAESU Linear 15p"));
#endif

  lcd.setCursor(3, 1);
  lcd.print(F(__DATE__));
  delay(1000);               //dirty delay to make information text readable
}

/*
    display TX activity
*/
void displayShowTX() {
  displayShowBand(1);                            // just call the function for band with the right parameter
}

// functions for other Hardware
/*
    display HW configuration on Serial
*/
void serialShowConfiguration() {
#if USE_CIV 
  DEBUG_PRINTLN(F("Icom CI-V"));
#ifdef FILTER_TO
  DEBUG_PRINT(F("FILTER_TO=")); DEBUG_PRINTLN(FILTER_TO, HEX);
#else
  DEBUG_PRINTLN(F("no FILTER_TO"));
#endif
#ifdef FILTER_FROM
  DEBUG_PRINT(F("FILTER_FROM=")); DEBUG_PRINTLN(FILTER_FROM, HEX);
#else
  DEBUG_PRINTLN(F("no FILTER_FROM"));
#endif
#endif   // USE_CIV
#if USE_YAESU_RS232
  DEBUG_PRINTLN(F("YAESU CAT RS232"));
#endif
#if USE_CIV || USE_YAESU_RS232
#if USE_SERIAL == -1
  DEBUG_PRINT(F("CAT SoftSerial on RXpin=")); DEBUG_PRINT(catRxPin); DEBUG_PRINT(F(" TXpin=")); DEBUG_PRINTLN(catTxPin);
#else
  DEBUG_PRINT(F("CAT HardwareSerial=")); DEBUG_PRINTLN(USE_SERIAL);
#endif
#endif
#if USE_ADC
  DEBUG_PRINTLN(F("ICOM ACC socket"));
  DEBUG_PRINT(F("read ACC from analog pin ")); DEBUG_PRINTLN(bandPin);
#endif
#if USE_YAESU_LINEAR
  DEBUG_PRINTLN(F("YAESU Linear 15p"));
  DEBUG_PRINT(F("read bitcode from pins ")); 
  DEBUG_PRINT(linearPin[0]); DEBUG_PRINT(' ');  // remark: a for loop for 4 items doesn't produce shorter code
  DEBUG_PRINT(linearPin[1]); DEBUG_PRINT(' ');
  DEBUG_PRINT(linearPin[2]); DEBUG_PRINT(' ');
  DEBUG_PRINTLN(linearPin[3]);
#endif
}

/*
    confirmation beep
*/
void toneStart() {
  tone(tonePin, 1200);
  delay(150);
  tone(tonePin, 1600);
  delay(150);
  tone(tonePin, 2000);
  delay(150);
  noTone(tonePin);
  // no delay at the end...
}

/*
    beep on key stroke
    currently implemented as non blocking
*/
void toneBeep() {
  if (setting.keyBeep > 0)
    tone(tonePin, 2000, setting.keyBeep);
}

/*
  Activates the antenna
  for future use: extend fo activateTX
*/
void activateRX(uint8_t newAntenna) {
  DEBUG_PRINT(F("D369 newAntenna=")); DEBUG_PRINTLN(newAntenna);
  for (size_t i = 0; i < noOfAntenna; i++) {
    if (i == newAntenna)
      digitalWrite(antenna[i].pin, relayON);
    else
      digitalWrite(antenna[i].pin, relayOFF);
  }
  displayShowAntenna();
}

/*
  map QRG to band index (array position)
  returns the band index 0 .. n) or 
           -1 if not found
*/
int8_t qrg2band(uint32_t qrg) {
  uint32_t search = qrg / 10;  // from Frequency*100 to Kilohertz
  for (size_t i = 0; i < noOfBand; i++) {
    if (search >= tblBandinformation[i].low && search <= tblBandinformation[i].high) return i;
  }
  return -1;       // not found in table
}

/*
   reset settings to an initial value
*/
void settingReset() {
  lcd.setCursor(0, 1);
  lcd.print(F(" HOLD TO RESET  "));
  delay(2000);   // this delay gives the user enough time to press both buttons
  if (digitalRead(leftPin) == LOW && digitalRead(rightPin) == LOW) {
    lcd.setCursor(0, 1);
    lcd.print(F(" RESETTING ...  "));
    for (uint8_t i = 0; i < noOfBand; i++) {
      setting.antennaRX[i] = 0;
    }
    //setting.antennaStart = 0;                  // no change of antennaStart so far, therefore comment out
    EEPROM.put(0, setting);
#ifdef ARDUINO_ARCH_ESP32
    EEPROM.commit();
#endif
    lcd.setCursor(0, 1);
    lcd.print(F(" Ready.         "));
    delay(1000);                                 // it will only wait for you to read the message
  }
}

/* *************************************************************************
   the Arduino setup()
 ************************************************************************* */
void setup() {
  // pin handling
  for (int8_t i = 0; i < (int8_t)noOfAntenna; i++) {
    pinMode(antenna[i].pin, OUTPUT);
    if (i == setting.antennaStart)               // set antennaStart to -1 if you don't want to start with an activated antenna
      digitalWrite(antenna[i].pin, relayON);
    else
      digitalWrite(antenna[i].pin, relayOFF);
  }
  pinMode(tonePin, OUTPUT);
  pinMode(leftPin, INPUT_PULLUP);
  pinMode(rightPin, INPUT_PULLUP);
  pinMode(pttPin, INPUT_PULLUP);                 // assumption ptt closes to GND

  mySerial.begin(9600);                          // CAT running at 9k6 - the serial port must be chosen defined in the configuration!
  Serial.begin(9600);                            // I know that's slow, but for simplicity I keep the debug output the same speed like the CAT interface
  DEBUG_PRINTLN(F("\r\nAntenna Switch"));
  DEBUG_PRINTLN(F(__DATE__));
  serialShowConfiguration();

  Wire.begin();                                  // needed for the I2C LCD
  lcd.begin();
  lcd.backlight();                               // backlight on
  lcd.createChar_P(0, cat);                      // special characters are only in PROGMEM - not in SRAM!
  lcd.createChar_P(3, memory);
  lcd.createChar_P(4, offline);

  displayShowStart();                            // greet the user
  settingReset();                                // check for reset
  lcd.clear();
#ifdef ARDUINO_ARCH_ESP32
  EEPROM.begin(1024);                            // As of 2022 the EEPROM emulation on the ESP32 is deprecated. It needs a separate call of .begin();
#endif
  EEPROM.get(0, setting);
  if (setting.antennaStart >= 0) {
    antennaRX = setting.antennaStart;
    activateRX(setting.antennaStart);  // assumption after startup there was no communication so far, therefore we activate a default value from setting
  }
  toneStart();                                   // startup beep

  // standard Screen
  lcd.setCursor(0, 0);
  lcd.print(F("QRG --.---- MHz  "));
  displayShowBand();
  displayShowAntenna();

#if USE_YAESU_RS232
  //yaesuBegin();      // initialise rig - to be tested
#endif

#if USE_YAESU_LINEAR
  yaesuLinearBegin();  // initialise pins to read
#endif

#if defined(__AVR_ATmega328P__) || defined(__AVR_ATmega2560__)
  wdt_enable(WDTO_8S);                   // activate watchdog
#endif
}

/* *************************************************************************
   the Arduino IDE loop()
 ************************************************************************* */
void loop() {
#if defined(__AVR_ATmega328P__) || defined(__AVR_ATmega2560__)
  wdt_reset();                                   // watchdog reset
#endif
  static bool previousPtt = LOW;                 // flag if ptt was pressed before
  static uint32_t previousButton = 0;            // timestamp when left was pressed before
  //static uint32_t previousRight = 0;           // timestamp when right was pressed before
  static bool changedSetting = false;            // flag if antenna setting was changed
  uint32_t currentMillis = millis();             // actual millis because we will do a lot of time related stuff and each millis() takes some time

  // read ptt
  if (digitalRead(pttPin) == LOW && previousPtt == false) {     // TX started
    DEBUG_PRINTLN(F("D510 TX on"));
    displayShowTX();                             // show TX (currently, this will replace the band name)
    previousPtt = true;
  }
  if (digitalRead(pttPin) == LOW && previousPtt == true) {      // TX ended
    DEBUG_PRINTLN(F("D415 TX off"));
    displayShowBand();                           // remove TX and show Band again
    previousPtt = false;
  }
  // read buttons
  if (digitalRead(leftPin) == LOW && digitalRead(pttPin) == HIGH && currentMillis - previousButton > readInterval) { // button press in RX
    if (antennaRX > 0)
      antennaRX--;
    else
      antennaRX = noOfAntenna - 1;
    DEBUG_PRINT(F("D525 new antennaRX=")); DEBUG_PRINTLN(antennaRX);
    activateRX(antennaRX);
    if (band >= 0) {
      setting.antennaRX[band] = antennaRX;
      changedSetting = true;
    }
    toneBeep();
    previousButton = currentMillis;
  }
  if (digitalRead(rightPin) == LOW && digitalRead(pttPin) == HIGH && currentMillis - previousButton > readInterval) { // button press in RX
    antennaRX = (antennaRX + 1) % noOfAntenna;
    DEBUG_PRINT(F("D536 new antennaRX=")); DEBUG_PRINTLN(antennaRX);
    activateRX(antennaRX);
    if (band >= 0) {
      setting.antennaRX[band] = antennaRX;
      changedSetting = true;
    }
    toneBeep();
    previousButton = currentMillis;
  }

  // read the QRG depending on the defined rig hardware
#if USE_CIV
  uint32_t QRGreaded = getFrequency();           // read CI-V interface
#endif
#if USE_ADC
  uint32_t QRGreaded = getFrequencyAdc();        // read band from ACC interface
#endif
#if USE_YAESU_RS232
  uint32_t QRGreaded = yaesuGetFrequency();      // read YAESU interface
  // yeasuTimer(); // retrigger timer?
#endif
#if USE_YAESU_LINEAR
  uint32_t QRGreaded = yaesuLinearGetFrequency();
#endif

  currentMillis = millis();                      // update timestamp just in case serial read took long
  if (QRGreaded > 0) {                           // new QRG
    int8_t previousBand = band;
    QRG = QRGreaded;
    band = qrg2band(QRG);                        // translate QRG to the band index
    DEBUG_PRINT(F("D566 QRG=")); DEBUG_PRINTLN(QRG);
    DEBUG_PRINT(F("D567 band=")); DEBUG_PRINTLN(band);
    if (band >= 0 && previousBand != band) {     // Band valid and differs
      size_t search = setting.antennaRX[band];
      DEBUG_PRINT(F("D557 search new antenna "));
      DEBUG_PRINTLN(search);
      if (search < noOfAntenna) {
        antennaRX = search;
        activateRX(search);
      }
    }
    displayShowQRG();
    indicatorQRGchange.on();                     // show QRG indicator for some time
    indicatorOffline.off();                      // MISSING tbc could be moved getFrequency()
  }
  // check if user setting needs to be stored into EEPROM
  if (changedSetting == true && (currentMillis - previousButton > 3000UL)) {
    DEBUG_PRINTLN(F("D583 persist data to EEPROM"));
    EEPROM.put(0, setting);                      // put only writes "changed" data, so it's no problem that we hand over the full struct
#ifdef ARDUINO_ARCH_ESP32
    EEPROM.commit();
#endif
    changedSetting = false;
    indicatorEEPROMchange.on();
  }
  if (currentMillis - previousQRGreceived > 30000UL) {
    indicatorOffline.on();                       // show "offline" indicator after some time
  }
  // call some objects to let their own time keeping running
  indicatorActivity.update();
  indicatorQRGchange.update();
  indicatorEEPROMchange.update();
  // indicatorOffline.update();                  // not needed for this indicator - offline has no timeout
}
