/* *************************************************************************
   Input Interface for ICOM - CI-V Protocol
   Reads the CI-V port
   ************************************************************************* */

#if USE_CIV
/*
   Decode a message from the ICOM rig.
   If it is a frequency message, return the frequency
   This keeps track of the current frequency and only
   returns a value != 0 when the frequency has changed
   reads message bytewise (one byte per call)
   returns 0 (no new QRG or still reading)
           or the new QRG (if reading was sucessfull)
*/
unsigned long getFrequency(void) {
  unsigned char c;                               // incoming character
  uint32_t t_QRG = 0;                            // QRG100 variable stores frequency in GMMMkkkH format - Frequ divided by 100
  uint8_t GHZ = 0, MHZ = 0, KHZ = 0, HZ = 0;
  static unsigned char buffget[16];              // read buffer
  static int8_t idx = 0;                         // index/counter of actual position
  static unsigned long lastfreq = 0;             // highest QRG100 in 32 bit: 4.294.967.295 --> 429 GHz (!)

  // If there's no input character return
  if (mySerial.available() <= 0) return (0);
  // If we haven't found an end of message, reset and try again.
  if (idx >= 15) {
    idx = 0;
    DEBUG_PRINTLN(F("D400 to long NOK"));
    return (0);
  }
  // Read the character
  c = buffget[idx] = mySerial.read();
  // Look for the 0xFE that starts the message
  if (idx == 0) {
    if (c == 0xFE) idx++;
    return (0);
  }
  // We've found the 0xFE, now look for the 0xFD at the end
  if (c != 0xFD) {
    idx++;
    return (0);
  }
#ifdef DEBUG
  for (int j = 0; j <= idx; j++) {
    if (buffget[j] < 10) Serial.print("0");
    Serial.print(buffget[j], HEX);
  }
  Serial.println();
#endif
  uint8_t len = idx;  // we destroy idx in next step but we need the received length later also
  idx = 0;
  // There is a complete message in buffget
  // If it is not a frequency message, ignore it
  // 0 is FE
  // 1 is FE
  // 2 is to address
  // 3 is from address --> this can be used to filter "wrong" devices on the same bus (set in configuration)
  if (buffget[0] != 0xFE || buffget[1] != 0xFE) {
    DEBUG_PRINTLN(F("D431 Startflag NOK"));
    return (0);
  }
  if (len != 9 && len != 10) {
    DEBUG_PRINTLN(F("D435 length NOK"));
    return (0);
  }
  if ((len == 9 && buffget[9] != 0xFD) || (len == 10 && buffget[10] != 0xFD) ) {
    DEBUG_PRINTLN(F("D439 endflag NOK"));
    return (0);
  }
#ifdef FILTER_TO
  if (buffget[2] != FILTER_TO) {
    DEBUG_PRINT(F("D444 ignore message to "));
    DEBUG_PRINTLN(buffget[2], HEX);
    return (0);
  }
#endif

#ifdef FILTER_FROM
  if (buffget[3] != FILTER_FROM) {
    DEBUG_PRINT(F("D452 ignore message from "));
    DEBUG_PRINTLN(buffget[3], HEX);
    return (0);
  }
#endif

  if ((buffget[4] != 0) && (buffget[4] != 3) &&  (buffget[4] != 5)) return (0); // the command 0, 3 or 5

  // decode the frequency and return it
  if (len == 10) {                               // do we have an long packet?
    GHZ = buffget[9];                            // 1GHz & 100Mhz
    GHZ = GHZ - (((GHZ / 16) * 6));              // explanation here: https://stackoverflow.com/questions/28133020/how-to-convert-bcd-to-decimal
    if (GHZ >= 100) return (0);
  }

  MHZ = buffget[8];                              // 10Mhz & 1Mhz
  MHZ = MHZ - (((MHZ / 16) * 6));                // Transform bytes ICOM CAT
  if (MHZ >= 100) return (0);

  KHZ = buffget[7];                              // 100Khz & 10KHz
  KHZ = KHZ - (((KHZ / 16) * 6));                // Transform bytes ICOM CAT
  if (KHZ >= 100) return (0);

  HZ = buffget[6];                               // 1Khz & 100Hz
  HZ = HZ - (((HZ / 16) * 6));                   // transform bytes ICOM CAT
  if (HZ >= 100) return (0);
  // buffget[5] is not readed as we don't store 10Hz & 1Hz

  t_QRG = ((GHZ * 1000000UL) + (MHZ * 10000UL) + (KHZ * 100) + (HZ * 1));

  previousQRGreceived = millis();                // we have received a valid QRG (so offline Indicator might get resetted)
  indicatorOffline.off();
  if (t_QRG == lastfreq) {
    indicatorActivity.on();
    return (0);              // but exit if it is the same
  }
  lastfreq = t_QRG;
  return (t_QRG);
}

#endif
